import os
from collections import defaultdict
from typing import Optional
import functools
import logging
from typing import NamedTuple
from urllib.parse import urlparse, parse_qsl

import pandas as pd

import arcticdb as adb


logger = logging.getLogger(__name__)

ENVIRONMENT = os.environ.get("ENVIRONMENT", "test")
ARCTIC_URL = os.environ.get(
    "TRADINGO_ARCTIC_URL",
    f"lmdb:///home/rory/dev/airflow/{ENVIRONMENT}/arctic.db",
)


class Symbol(NamedTuple):

    library: str
    symbol: str
    kwargs: dict


def lib_provider(**libs):

    def decorator(func):

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            arctic = adb.Arctic(ARCTIC_URL)
            libs_ = {
                k: arctic.get_library(v, create_if_missing=True)
                for k, v in libs.items()
            }
            return func(*args, **libs_, **kwargs)

        return wrapper

    return decorator


def parse_symbol(symbol, kwargs, symbol_prefix="", symbol_postfix=""):
    string_kwargs = {k: str(v) for k, v in kwargs.items()}
    symbol = symbol.format(**string_kwargs)
    parsed_symbol = urlparse(symbol)
    lib, sym = parsed_symbol.path.split("/")
    kwargs = dict(parse_qsl(parsed_symbol.query))
    symbol_prefix = symbol_prefix.format(**string_kwargs)
    symbol_postfix = symbol_postfix.format(**string_kwargs)
    for key, value in kwargs.items():
        if key == "as_of":
            try:
                kwargs[key] = int(value)
            except TypeError:
                continue
    return Symbol(lib, symbol_prefix + sym + symbol_postfix, kwargs)


def symbol_provider(
    symbol_prefix="",
    no_date=False,
    **symbols,
):

    def decorator(func):

        @functools.wraps(func)
        def wrapper(
            *args,
            start_date=None,
            end_date=None,
            arctic: Optional[adb.Arctic] = None,
            **kwargs,
        ):

            arctic = arctic or adb.Arctic(ARCTIC_URL)
            orig_symbol_data = {}
            requested_symbols = symbols.copy()

            for symbol in symbols:
                if symbol in kwargs and isinstance(
                    kwargs[symbol], (pd.DataFrame, pd.Series)
                ):
                    orig_symbol_data[symbol] = kwargs.pop(symbol)
                if symbol in kwargs and isinstance(kwargs[symbol], str):
                    requested_symbols[symbol] = kwargs[symbol]
                if symbol in kwargs and kwargs[symbol] is None:
                    requested_symbols.pop(symbol)

            symbols_data = {
                k: (
                    arctic.get_library(
                        parse_symbol(v, kwargs, symbol_prefix=symbol_prefix).library,
                        create_if_missing=True,
                    )
                    .read(
                        parse_symbol(v, kwargs, symbol_prefix=symbol_prefix).symbol,
                        date_range=(
                            None
                            if no_date
                            else (
                                pd.Timestamp(start_date) if start_date else None,
                                pd.Timestamp(end_date) if end_date else None,
                            )
                        ),
                        **parse_symbol(v, kwargs, symbol_prefix=symbol_prefix).kwargs,
                    )
                    .data
                )
                for k, v in requested_symbols.items()
                if k not in orig_symbol_data
            }
            kwargs.update(symbols_data)
            kwargs.update(orig_symbol_data)

            logger.info("Providing %s symbols from %s", symbols_data.keys(), arctic)

            return func(
                *args, **kwargs, arctic=arctic, start_date=start_date, end_date=end_date
            )

        return wrapper

    return decorator


def symbol_publisher(
    *symbols,
    symbol_prefix="",
    symbol_postfix="",
    astype=None,
    template=None,
    library_options=None,
    write_pickle=False,
):

    def decorator(func):

        @functools.wraps(func)
        def wrapper(
            *args,
            dry_run=False,
            arctic: Optional[adb.Arctic] = None,
            snapshot: Optional[str] = None,
            clean: bool = False,
            **kwargs,
        ):

            arctic = arctic or adb.Arctic(ARCTIC_URL)
            out = func(*args, arctic=arctic, **kwargs)
            logger.info("Publishing %s to %s", symbols or template, arctic)

            # yf kws:
            #    kwdout = out.pop(-1)
            #
            if template:
                out, symbols_ = list(zip(*out))
                symbols_ = [template.format(*s) for s in symbols_]
            else:
                symbols_ = symbols

            libraries = defaultdict(dict)

            for data, symbol in zip(out, symbols_):

                if astype:
                    data = data.astype(
                        astype
                        if not isinstance(astype, dict)
                        else {k: v for k, v in astype.items() if k in data.columns}
                    )

                logger.info(
                    "writing symbol=%s rows=%s",
                    parse_symbol(symbol, kwargs, symbol_prefix, symbol_postfix),
                    len(data.index),
                )

                lib, sym, params = parse_symbol(
                    symbol,
                    kwargs,
                    symbol_prefix=symbol_prefix,
                    symbol_postfix=symbol_postfix,
                )

                if not dry_run:
                    lib = arctic.get_library(
                        lib, create_if_missing=True, library_options=library_options
                    )
                    if isinstance(data.index, pd.DatetimeIndex):
                        if clean:
                            lib.delete(sym)
                        result = lib.update(
                            sym,
                            data,
                            upsert=True,
                            date_range=(data.index[0], data.index[-1]),
                            **params,
                        )
                    elif write_pickle:
                        result = lib.write_pickle(sym, data, **params)

                    else:
                        result = lib.write(sym, data, **params)

                    libraries[lib.name][result.symbol] = result.version

            if not dry_run and snapshot:
                for lib_name, versions in libraries.items():
                    logging.info(
                        "Snapshotting %s for %s %s", lib_name, snapshot, versions
                    )
                    lib = arctic.get_library(lib_name)
                    if snapshot in lib.list_snapshots():
                        lib.delete_snapshot(snapshot)
                    lib.snapshot(snapshot_name=snapshot, versions=versions)

            if dry_run:
                return pd.concat(out, keys=symbols_, axis=1)

            return None

        wrapper.published_fields = symbols
        return wrapper

    return decorator



from __future__ import annotations
import logging
from enum import Enum
import importlib
import functools
import argparse
from json.decoder import JSONDecodeError
import pathlib
import json
from typing import Callable, Iterable
from arcticdb import Arctic

import pandas as pd

from tradingo.symbols import ARCTIC_URL

SIGNAL_KEY = "{0}.{1}".format

DEFAULT_STAGE = "raw.shares"


def resolve_config(
    config: str | pathlib.Path,
):
    try:
        return json.loads(config)
    except (JSONDecodeError, TypeError) as _:
        return json.loads(pathlib.Path(config).read_text())


def cli_app():
    app = argparse.ArgumentParser("tradingo")

    app.add_argument("--config", type=resolve_config)
    app.add_argument("--task", required=True)
    app.add_argument("--with-deps", action="store_true")
    app.add_argument("--start-date", type=pd.Timestamp)
    app.add_argument("--end-date", type=pd.Timestamp)
    app.add_argument("--force-rerun", action="store_true")
    app.add_argument("--arctic-uri", default=ARCTIC_URL)
    return app


def task_resolver(func):

    functools.wraps(func)

    def wrapper(*args, global_tasks, **kwargs):

        tasks_out = func(*args, **kwargs)
        global_tasks.update(tasks_out)

        for task in tasks_out.values():
            task.resolve_dependencies(global_tasks)

        return tasks_out

    return wrapper


class TaskState(Enum):

    PENDING = "PENDING"
    FAILED = "FAILED"
    SUCCESS = "SUCCESS"


class Task:

    def __init__(
        self,
        function,
        task_args,
        task_kwargs,
        dependencies: Iterable[str] = (),
    ):
        self._function = function
        self.task_args = task_args
        self.task_kwargs = task_kwargs
        self._dependencies = list(dependencies)
        self._resolved_dependencies = []
        self.state = TaskState.PENDING

    def __repr__(self):
        return (
            f"Task(function='{self._function}',"
            f" task_args={self.task_args},"
            f" task_kwargs={self.task_kwargs},"
            f" dependcies={self._dependencies}, "
            ")"
        )

    @property
    def function(self) -> Callable:
        module, function_name = self._function.rsplit(".", maxsplit=1)
        return getattr(importlib.import_module(module), function_name)

    def run(self, *args, run_dependencies=False, force_rerun=False, **kwargs):

        if run_dependencies:

            for dependency in self.dependencies:

                dependency.run(
                    run_dependencies=run_dependencies,
                    force_rerun=force_rerun,
                )

        state = self.state
        try:
            if self.state == TaskState.PENDING or force_rerun:
                state = TaskState.FAILED
                print(f"Running {self}")
                self.function(*self.task_args, *args, **self.task_kwargs, **kwargs)
                self.state = state = TaskState.SUCCESS
        finally:
            self.state = state

    def add_dependencies(self, *dependency):
        self._dependencies.extend(dependency)

    def resolve_dependencies(self, tasks: dict[str, Task]):
        self._resolved_dependencies.extend(
            tasks[dep_name] for dep_name in self._dependencies
        )

    @property
    def dependencies(self) -> list[Task]:
        return self._resolved_dependencies

    @property
    def dependency_names(self) -> list[str]:
        return self._dependencies


@task_resolver
def collect_signal_tasks(
    signals: dict[str, dict],
    universe: str,
    provider: str,
    start_date: pd.Timestamp,
    end_date: pd.Timestamp,
):

    tasks = {
        SIGNAL_KEY(universe, signal): Task(
            config["function"],
            config.get("args", []),
            {
                **config.get("kwargs", {}),
                "start_date": start_date,
                "end_date": end_date,
                "universe": universe,
                "provider": provider,
            },
            [
                f"{universe}.sample",
                f"{universe}.vol",
                # f"{universe}.ivol",
                *(SIGNAL_KEY(universe, k) for k in config.get("depends_on", [])),
            ],
        )
        for signal, config in signals.items()
    }

    return tasks


@task_resolver
def collect_sample_tasks(
    universes: dict[str, dict],
    start_date: pd.Timestamp,
    sample_start_date: pd.Timestamp,
    end_date: pd.Timestamp,
):

    tasks = {}

    for universe, config in universes.items():

        # TODO: start,end date
        #
        #
        provider = config["provider"]

        tasks[f"{universe}.instruments"] = Task(
            "tradingo.sampling.download_instruments",
            [],
            {
                "html": config.get("html"),
                "file": config.get("file"),
                "tickers": config.get("tickers"),
                "epics": config.get("epics"),
                "index_col": config["index_col"],
                "universe": universe,
            },
        )
        tasks[f"{universe}.sample"] = Task(
            config.get("function", "tradingo.sampling.sample_equity"),
            [],
            {
                "start_date": sample_start_date,
                "end_date": end_date,
                "provider": provider,
                "interval": config.get("interval", "1d"),
                "universe": universe,
                "periods": config.get("periods"),
            },
            [f"{universe}.instruments"],
        )
        tasks[f"{universe}.vol"] = Task(
            "tradingo.signals.vol",
            [],
            {
                "start_date": start_date,
                "end_date": end_date,
                "provider": provider,
                # TODO: speeds
                "speeds": config["volatility"]["speeds"],
                "universe": universe,
                "close": config["volatility"].get("field", "adj_close"),
            },
            [f"{universe}.sample"],
        )

        if config["volatility"].get("ivol", True):

            tasks[f"{universe}.ivol"] = Task(
                "tradingo.portfolio.instrument_ivol",
                [],
                {
                    "start_date": None,
                    "end_date": None,
                    "provider": config["provider"],
                    "interval": config.get("interval"),
                    "universe": universe,
                    "close": config["volatility"].get("field", "adj_close"),
                },
                [f"{universe}.sample"],
            )

    return tasks


def build_graph(
    config,
    start_date,
    end_date,
    sample_start_date=None,
    snapshot_template=None,
) -> dict[str, Task]:

    sample_start_date = sample_start_date or start_date

    global_tasks = {}

    sample_tasks = collect_sample_tasks(
        config["universe"],
        global_tasks=global_tasks,
        start_date=start_date,
        sample_start_date=sample_start_date,
        end_date=end_date,
    )

    for portfolio_name, portfolio_config in config["portfolio"].items():

        universe = portfolio_config["universe"]
        provider = portfolio_config["provider"]

        portfolio_config["kwargs"].setdefault("start_date", start_date)
        portfolio_config["kwargs"].setdefault("end_date", end_date)
        portfolio_config["kwargs"].setdefault("name", portfolio_name)
        portfolio_config["kwargs"].setdefault("provider", provider)
        portfolio_config["kwargs"].setdefault("universe", universe)
        portfolio_config["kwargs"].setdefault("snapshot", snapshot_template)

        portfolio_task = Task(
            portfolio_config["function"],
            portfolio_config.get("args", []),
            portfolio_config.get("kwargs", {}),
            dependencies=[
                f"{universe}.sample",
                f"{universe}.vol",
                # f"{universe}.ivol",
                *(
                    SIGNAL_KEY(universe, sig)
                    for sig in portfolio_config["kwargs"].get("signal_weights", {})
                ),
            ],
        )

        signals = collect_signal_tasks(
            config["signal_configs"],
            portfolio_config["universe"],
            portfolio_config["provider"],
            global_tasks=global_tasks,
            start_date=start_date,
            end_date=end_date,
        )

        portfolio_task.resolve_dependencies(global_tasks)

        global_tasks[portfolio_name] = portfolio_task

        backtest_kwargs = portfolio_config.get("backtest", {"stage": DEFAULT_STAGE})
        trade_kwargs = portfolio_config.get("trades", {"stage": DEFAULT_STAGE})

        backtest = global_tasks[f"{portfolio_name}.backtest"] = Task(
            "tradingo.backtest.backtest",
            task_args=[],
            task_kwargs={
                "start_date": start_date,
                "end_date": end_date,
                "name": portfolio_name,
                "provider": provider,
                "universe": universe,
                **backtest_kwargs,
            },
            dependencies=[portfolio_name],
        )

        trades = global_tasks[f"{portfolio_name}.trades"] = Task(
            "tradingo.portfolio.calculate_trades",
            task_args=[],
            task_kwargs={
                "start_date": start_date,
                "end_date": end_date,
                "name": portfolio_name,
                "provider": provider,
                "universe": universe,
                **trade_kwargs,
            },
            dependencies=[portfolio_name],
        )

        downstream = global_tasks[f"{portfolio_name}.downstream"] = Task(
            "tradingo.engine.adjust_position_sizes",
            task_args=[],
            task_kwargs={
                "portfolio_name": portfolio_name,
                "name": config["name"],
                "provider": provider,
                "universe": universe,
                "stage": backtest_kwargs["stage"],
                "arctic_uri": ARCTIC_URL,
            },
            dependencies=[portfolio_name],
        )

        downstream.resolve_dependencies(global_tasks)
        backtest.resolve_dependencies(global_tasks)
        trades.resolve_dependencies(global_tasks)

    return global_tasks


def serialise_dag(graph: dict[str, Task]):

    dag_state = pathlib.Path.home() / ".tradingo/dag-state.json"

    dag_state.parent.mkdir(parents=True, exist_ok=True)
    dag_state.write_text(
        json.dumps({k: v.state.value for k, v in graph.items()}, indent=2)
    )


def update_dag(graph: dict[str, Task]):

    dag_state = pathlib.Path.home() / ".tradingo/dag-state.json"

    if not dag_state.exists():

        return

    else:
        dag_state = json.loads(dag_state.read_text())

        for k, v in dag_state.items():

            if k not in graph:
                continue
            state = TaskState[v]

            graph[k].state = state if state == TaskState.SUCCESS else TaskState.PENDING


def main():

    args = cli_app().parse_args()

    graph = build_graph(args.config, args.start_date, args.end_date)
    arctic = Arctic(args.arctic_uri)

    update_dag(graph)

    task = graph[args.task]

    try:
        task.run(
            run_dependencies=args.with_deps, force_rerun=args.force_rerun, arctic=arctic
        )
    finally:
        serialise_dag(graph)


if __name__ == "__main__":

    logging.getLogger("tradingo").setLevel(logging.INFO)
    main()



import logging
import operator
import numba
import numpy as np
import math
import functools

import pandas as pd

from pandas._libs.tslibs import IncompatibleFrequency
import pandas_market_calendars as pmc

from tradingo.symbols import symbol_provider, symbol_publisher


logger = logging.getLogger(__name__)


@symbol_provider(
    close="prices/adj_close",
    symbol_prefix="{provider}.{universe}.",
)
@symbol_publisher(
    template="signals/vol_{0}",
    symbol_prefix="{provider}.{universe}.",
)
def vol(
    speeds,
    close: pd.DataFrame,
    **kwargs,
):
    returns = np.log(close / close.shift())
    return tuple(
        (
            returns.ewm(halflife=speed).std(),
            (speed,),
        )
        for speed in speeds
    )


@symbol_provider(close="prices/adj_close", symbol_prefix="{provider}.{universe}.")
@symbol_publisher(
    "signals/{signal_name}",
    symbol_prefix="{provider}.{universe}.",
)
def ewmac_signal(
    close: pd.DataFrame,
    speed1: int,
    speed2: int,
    provider: str,
    config_name: str,
    model_name="ewmac",
    signal_name="ewmac_{speed1}_{speed2}",
    **kwargs,
):

    logger.info(
        "Running %s model=%s signal=%s with %s",
        config_name,
        model_name,
        signal_name,
        provider,
    )

    returns = np.log(close / close.shift())

    return (
        (
            close.ewm(halflife=speed2, min_periods=speed2).mean()
            - close.ewm(halflife=speed1, min_periods=speed1).mean()
        ),
    )


@symbol_publisher(
    "signals/{signal}.scaled",
    symbol_prefix="{provider}.{universe}.{model_name}.",
)
@symbol_provider(
    signal="signals/{signal}",
    symbol_prefix="{provider}.{universe}.{model_name}.",
)
def scaled(signal, scale: float, **kwargs):
    return ((signal / signal.abs().max()) * scale,)


@symbol_publisher(
    "signals/{signal}.capped",
    symbol_prefix="{provider}.{universe}.{model_name}.",
)
@symbol_provider(
    signal="signals/{signal}",
    symbol_prefix="{provider}.{universe}.{model_name}.",
)
def capped(signal: pd.Series, cap: float, **kwargs):
    signal[signal.abs() >= cap] = np.sign(signal) * cap
    return (signal,)


@numba.jit
def _linear_buffer(signal: np.ndarray, thresholds: np.ndarray):

    res: np.ndarray = np.copy(signal)
    lower = signal - thresholds
    upper = signal + thresholds
    res[0, :] = signal[0, :]

    for i in range(1, res.shape[0]):
        for j in range(signal.shape[1]):
            trade_in = (
                math.isnan(res[i - 1, j]) or res[i - 1, j] == 0.0
            ) and not math.isnan(signal[i, j])
            if trade_in:
                res[i, j] = signal[i, j]
            elif res[i - 1, j] < lower[i, j]:
                res[i, j] = lower[i, j]
            elif res[i - 1, j] > upper[i, j]:
                res[i, j] = upper[i, j]
            else:
                res[i, j] = res[i - 1, j]
    if signal.ndim == 1:
        return np.squeeze(res)
    return res


@symbol_publisher(
    "{library}/{signal}.buffered", symbol_prefix="{provider}.{universe}.{model_name}."
)
@symbol_provider(
    signal="{library}/{signal}",
    symbol_prefix="{provider}.{universe}.{model_name}.",
)
def buffered(signal: pd.Series | pd.DataFrame, buffer_width, **kwargs):

    signal = signal.ffill().fillna(0.0)

    thresholds = signal.mul(buffer_width).abs()

    buffered = _linear_buffer(signal.to_numpy(), thresholds.to_numpy())

    return (pd.DataFrame(buffered, index=signal.index, columns=signal.columns),)


@symbol_publisher(
    "signals/intraday_momentum",
    "signals/intraday_momentum.z_score",
    "signals/intraday_momentum.short_vol",
    "signals/intraday_momentum.long_vol",
    "signals/intraday_momentum.previous_close_px",
    symbol_prefix="{provider}.{universe}.",
)
@symbol_provider(
    ask_close="prices/ask.close",
    bid_close="prices/bid.close",
    symbol_prefix="{provider}.{universe}.",
)
def intraday_momentum(
    ask_close,
    bid_close,
    calendar="NYSE",
    frequency="15min",
    long_vol=64,
    short_vol=6,
    cap=2,
    threshold=1,
    close_offset_periods: int = 0,
    monotonic: bool = True,
    incremental: int = 0,
    only_with_close: bool = True,
    **kwargs,
):

    close = (ask_close + bid_close) / 2

    cal = pmc.get_calendar(calendar)
    schedule = cal.schedule(start_date=close.index[0], end_date=close.index[-1])
    trading_index = pmc.date_range(schedule, frequency=frequency).intersection(
        close.index
    )
    close = close.reindex(trading_index)

    close_px = close.groupby(close.index.date).last()

    close_px.index = pd.to_datetime(close_px.index).tz_localize("utc")

    previous_close_px = close_px.shift()

    long_vol = (
        previous_close_px.pct_change()
        .ewm(long_vol, min_periods=long_vol)
        .std()
        .reindex(close.index, method="ffill")
    )
    short_vol = (
        previous_close_px.pct_change()
        .ewm(short_vol, min_periods=short_vol)
        .std()
        .rolling(10)
        .quantile(q=0.75)
        .reindex(close.index, method="ffill")
    )

    previous_close_px = previous_close_px.reindex(
        close.index,
        method="ffill",
    )
    close_px = close_px.reindex(close.index, method="ffill")

    z_score = ((close - previous_close_px) / previous_close_px) / long_vol

    signal = z_score.copy()

    # apply caps and threshold
    #
    signal = (
        signal.where(signal.abs() > threshold, 0)
        .where(signal.abs() < cap, np.sign(signal) * cap)
        .groupby(signal.index.date)
        .ffill()
        .fillna(0)
    )

    signal = signal.where(
        signal.index.to_series().dt.date.eq(signal.index.to_series().dt.date.shift()),
        0,  # make first reading 0.0 position
    ).where(
        np.sign(signal.shift()).eq(np.sign(signal)), 0  # make changes in sign 0
    )

    def dynamic_floor(series, shift=0):
        return series.groupby(series.index.date).transform(
            lambda i: i.where(
                ~functools.reduce(
                    operator.and_,
                    (
                        (np.sign(i.shift(s)).ne(0) & np.sign(i.shift(s)).ne(np.sign(i)))
                        for s in range(1, shift)
                    ),
                ),
                0,
            )
        )

    signal = dynamic_floor(signal, 5)

    # allocate 10_000 notional in unit vol space
    gearing = 10_000 / previous_close_px
    signal = (gearing * signal) / (short_vol * np.sqrt(252))
    # signal = signal.abs().groupby(signal.index.date).cummax() * np.sign(z_score)
    # signal closes position at close time

    def periods_before_close(n):

        return functools.reduce(
            pd.DatetimeIndex.union,
            (
                schedule.market_close
                - (i * pd.tseries.frequencies.to_offset(frequency))
                for i in range(1, n + 1)
            ),
            pd.DatetimeIndex(schedule.market_close),
        )

    close_at = periods_before_close(n=close_offset_periods)

    if monotonic:

        signal_cumabsmax = signal.abs().groupby(signal.index.date).cummax()
        is_long = np.sign(signal).groupby(signal.index.date).cummax()
        is_short = np.sign(signal).groupby(signal.index.date).cummin()
        direction = is_long.where(is_long > 0, is_short)
        signal = signal_cumabsmax * np.sign(direction)

    if incremental:

        no_trade_at = periods_before_close(n=incremental)
        signal.loc[signal.index.isin(no_trade_at)] = 0.0
        signal = signal.groupby(signal.index.date).cumsum() * 1

    if only_with_close:

        has_close = (
            close.groupby(close.index.date)
            .apply(
                lambda i: i[
                    i.index.tz_convert(cal.close_time.tzinfo).time == cal.close_time
                ]
                .notna()
                .reindex(i.index, method="bfill")
            )
            .droplevel(0)
        )
        signal = signal.where(has_close, 0)

    signal.loc[signal.index.isin(close_at)] = 0.0

    # signal = signal.where((close > previous_close_px) & np.sign(signal) > 0, 0).where(
    #     (close < previous_close_px) & np.sign(signal) < 0, 0
    # )

    return (
        signal,
        z_score,
        short_vol,
        long_vol,
        previous_close_px,
    )



import dateutil.tz
from pandas.io.pytables import import_optional_dependency
from tradingo.symbols import symbol_provider, symbol_publisher
from typing import Literal, Optional

from trading_ig.rest import IGService, ApiExceededException
from trading_ig.config import config

from tenacity import Retrying, wait_exponential, retry_if_exception_type

from arcticdb import LibraryOptions
from yfinance import Ticker

import pandas as pd
import numpy as np


FUTURES_FIELDS = [
    "expiration",
    "price",
]


OPTION_FIELDS = [
    "expiration",
    "option_type",
    "strike",
    "open_interest",
    "volume",
    "theoretical_price",
    "last_trade_price",
    "tick",
    "bid",
    "bid_size",
    "ask",
    "ask_size",
    "open",
    "high",
    "low",
    "prev_close",
    "change",
    "change_percent",
    "implied_volatility",
    "delta",
    "gamma",
    "theta",
    "vega",
    "rho",
    "last_trade_timestamp",
    "dte",
]


Provider = Literal[
    "alpha_vantage",
    "cboe",
    "fmp",
    "intrinio",
    "polygon",
    "tiingo",
    "tmx",
    "tradier",
    "yfinance",
]


def get_ig_service() -> IGService:

    retryer = Retrying(
        wait=wait_exponential(),
        retry=retry_if_exception_type(ApiExceededException),
    )

    service = IGService(
        username=config.username,
        password=config.password,
        api_key=config.api_key,
        acc_type=config.acc_type,
        use_rate_limiter=True,
        retryer=retryer,
    )

    service.create_session()
    return service


@symbol_publisher("instruments/{universe}", write_pickle=True)
def download_instruments(
    index_col: str,
    *,
    html: Optional[str] = None,
    file: Optional[str] = None,
    tickers: Optional[list[str]] = None,
    epics: Optional[list[str]] = None,
    **kwargs,
):

    if file:
        return (
            pd.read_csv(
                file,
                index_col=index_col,
            ).rename_axis("Symbol"),
        )
    if html:
        return (pd.read_html(html)[0].set_index(index_col).rename_axis("Symbol"),)

    if tickers:

        return (
            (
                pd.DataFrame({t: Ticker(t).get_info() for t in tickers})
                .transpose()
                .rename_axis("Symbol")
            ),
        )

    if epics:

        service = get_ig_service()

        return (
            pd.DataFrame((service.fetch_market_by_epic(e)["instrument"] for e in epics))
            .set_index("epic")
            .rename_axis("Symbol", axis=0),
        )
    raise ValueError(file)


@symbol_provider(instruments="instruments/{universe}", no_date=True)
@symbol_publisher(
    template="prices/{0}.{1}",
    symbol_prefix="{provider}.{universe}.",
)
def sample_ig_instruments(
    instruments: pd.DataFrame,
    end_date: pd.Timestamp,
    start_date: pd.Timestamp,
    interval: str,
    **kwargs,
):

    service = get_ig_service()

    def get_data(symbol):
        try:
            return service.fetch_historical_prices_by_epic(
                symbol,
                end_date=pd.Timestamp(end_date)
                .tz_convert(dateutil.tz.tzlocal())
                .tz_localize(None)
                .isoformat(),
                start_date=(pd.Timestamp(start_date) + pd.Timedelta(seconds=1))
                .tz_convert(dateutil.tz.tzlocal())
                .tz_localize(None)
                .isoformat(),
                resolution=interval,
                wait=0,
            )["prices"]
        except Exception as ex:
            if ex.args[0] == "Historical price data not found":
                return pd.DataFrame(
                    np.nan,
                    columns=pd.MultiIndex.from_tuples(
                        (
                            ("Open", "bid"),
                            ("Open", "ask"),
                            ("High", "bid"),
                            ("High", "ask"),
                            ("Low", "bid"),
                            ("Low", "ask"),
                        )
                    ),
                    index=[],
                )  # TODO:
            raise ex

    result = pd.concat(
        ((get_data(symbol) for symbol in instruments.index.to_list())),
        axis=1,
        keys=instruments.index.to_list(),
    ).reorder_levels([1, 2, 0], axis=1)
    result.index = result.index.tz_localize(dateutil.tz.tzlocal()).tz_convert("utc")
    return (
        (result["bid"]["Open"], ("bid", "open")),
        (result["bid"]["High"], ("bid", "high")),
        (result["bid"]["Low"], ("bid", "low")),
        (result["bid"]["Close"], ("bid", "close")),
        (result["ask"]["Open"], ("ask", "open")),
        (result["ask"]["High"], ("ask", "high")),
        (result["ask"]["Low"], ("ask", "low")),
        (result["ask"]["Close"], ("ask", "close")),
        (((result["ask"]["Open"] + result["bid"]["Open"]) / 2), ("mid", "open")),
        (((result["ask"]["High"] + result["bid"]["High"]) / 2), ("mid", "high")),
        (((result["ask"]["Low"] + result["bid"]["Low"]) / 2), ("mid", "low")),
        (((result["ask"]["Close"] + result["bid"]["Close"]) / 2), ("mid", "close")),
        # (result["last"]["Open"], ("last", "open")),
        # (result["last"]["High"], ("last", "high")),
        # (result["last"]["Low"], ("last", "low")),
        # (result["last"]["Close"], ("last", "close")),
        # (result["last"]["Volume"], ("last", "volume")),
    )


@symbol_provider(instruments="instruments/{universe}", no_date=True)
@symbol_publisher(
    "prices/open",
    "prices/high",
    "prices/low",
    "prices/close",
    "prices/adj_close",
    "prices/volume",
    "prices/dividend",
    "prices/split_ratio",
    astype=float,
    symbol_prefix="{provider}.{universe}.",
)
def sample_equity(
    instruments: pd.DataFrame,
    start_date: str,
    end_date: str,
    provider: Provider,
    interval: str = "1d",
    **kwargs,
):
    from openbb import obb

    data = obb.equity.price.historical(  # type: ignore
        instruments.index.to_list(),
        start_date=start_date,
        end_date=end_date,
        provider=provider,
        interval=interval,
    ).to_dataframe()

    data.index = pd.to_datetime(data.index)

    close = data.pivot(columns=["symbol"], values="close")
    close.index = pd.to_datetime(close.index)

    return (
        data.pivot(columns=["symbol"], values="open"),
        data.pivot(columns=["symbol"], values="high"),
        data.pivot(columns=["symbol"], values="low"),
        close,
        100 * (1 + close.pct_change()).cumprod(),
        data.pivot(columns=["symbol"], values="volume"),
        data.pivot(columns=["symbol"], values="dividend"),
        data.pivot(columns=["symbol"], values="split_ratio"),
    )


@symbol_publisher(
    template="options/{0}.{1}",
    symbol_prefix="{provider}.",
    library_options=LibraryOptions(dynamic_schema=True, dedup=True),
)
def sample_options(
    universe: list[str], start_date: str, end_date: str, provider: Provider, **kwargs
):
    from openbb import obb

    out = []

    data = (
        (
            symbol,
            obb.derivatives.options.chains(symbol)  # type: ignore
            .to_dataframe()
            .replace({"call": 1, "put": -1})
            .astype({"expiration": "datetime64[ns]"})
            .assign(timestamp=end_date)
            # .set_index(["timestamp", "expiration", "strike", "option_type"])
            .set_index(["timestamp", "contract_symbol"]).unstack(level=1),
        )
        for symbol in universe
    )

    for symbol, df in data:
        df.index = pd.to_datetime(df.index)
        out.extend((df[field], (symbol, field)) for field in OPTION_FIELDS)

    return out


@symbol_publisher(
    template="futures/{0}.{1}",
    symbol_prefix="{provider}.",
    library_options=LibraryOptions(dynamic_schema=True, dedup=True),
)
def sample_futures(
    universe: list[str], start_date: str, end_date: str, provider: Provider, **kwargs
):
    from openbb import obb

    out = []

    data = (
        (
            symbol,
            obb.derivatives.futures.curve(symbol)
            .to_dataframe()
            .assign(timestamp=end_date)
            .set_index(["timestamp", "symbol"])
            .unstack("symbol"),
        )
        for symbol in universe
    )

    for symbol, df in data:
        df.index = pd.to_datetime(df.index)
        out.extend((df[field], (symbol, field)) for field in FUTURES_FIELDS)

    return out



import json
import os
import pathlib
import pandas as pd


def get_instruments(config, key="equity") -> pd.DataFrame:
    if "file" in config[key]:
        return pd.read_csv(
            config[key]["file"],
            index_col=config[key]["index_col"],
        ).rename_axis("Symbol")
    if "html" in config[key]:
        return pd.read_html(config[key]["html"], index_col=config[key]["index_col"])[0]
    raise ValueError(config[key])


def with_instrument_details(
    dataframe: pd.DataFrame,
    instruments: pd.DataFrame,
    fields: list[str],
):
    """Add instrument details to column index"""
    return (
        dataframe.transpose()
        .rename_axis("Symbol")
        .merge(instruments[fields], left_index=True, right_index=True)
        .reset_index()
        .set_index([*fields, "Symbol"])
        .sort_index()
        .transpose()
    ).dropna()


def null_instruments(symbols):
    return pd.DataFrame(
        data="",
        index=symbols,
        columns=[
            "Name",
            "SEDOL",
            "ISIN",
            "CUSIP",
            "Incept. Date",
            "Gross Expense Ratio (%)",
            "Net Expense Ratio (%)",
            "Net Assets (USD)",
            "Net Assets as of",
            "Asset Class",
            "Sub Asset Class",
            "Region",
            "Market",
            "Location",
            "Investment Style",
            "Key Facts",
            "Avg. Annual Return: NAV Quarterly",
            "Avg. Annual Return: Price Quarterly",
            "Avg. Annual Return: NAV Monthly",
            "Avg. Annual Return: Price Monthly",
            "Yield",
            "Fixed Income Characteristics",
            "Sustainability Characteristics (MSCI ESG Fund Ratings)",
        ],
    )


def buffer(
    series: pd.Series,
):
    pass



import pytest
import pandas as pd
import numpy as np
import pandas_market_calendars as pmc
from tradingo.api import Tradingo

from tradingo import signals


UNIVERSE = "test-universe"
PROVIDER = "test-provider"


def test_buffer_signal(tradingo: Tradingo):

    signals.buffered(
        signal=tradingo.portfolio.model.raw.shares(),
        thresholds=tradingo.portfolio.model.raw.shares() * 0.1,
        dry_run=True,
        start_date="2023-01-01",
        end_date="2024-05-31",
        config_name="test",
        model_name="model",
        library="portfolio",
        buffer_width=0.5,
        provider="yfinance",
        universe="etfs",
        arctic=tradingo,
    )


@pytest.mark.parametrize(
    "ask_close, bid_close, close_offset_periods",
    [
        (
            pd.DataFrame(
                np.cumprod(1 + np.random.normal(0, 0.009, 167905)),
                index=pd.bdate_range(
                    start="2020-01-01 00:00:00+00:00",
                    end="2024-10-15 18:30:00+00:00",
                    freq=pd.offsets.Minute(15),
                ),
                columns=["ABCD"],
            ),
            pd.DataFrame(
                np.cumprod(1 + np.random.normal(0, 0.009, 167905)),
                index=pd.bdate_range(
                    start="2020-01-01 00:00:00+00:00",
                    end="2024-10-15 18:30:00+00:00",
                    freq=pd.offsets.Minute(15),
                ),
                columns=["ABCD"],
            ),
            0,
        ),
        (
            pd.DataFrame(
                np.cumprod(1 + np.random.normal(0, 0.009, 167905)),
                index=pd.bdate_range(
                    start="2020-01-01 00:00:00+00:00",
                    end="2024-10-15 18:30:00+00:00",
                    freq=pd.offsets.Minute(15),
                ),
                columns=["ABCD"],
            ),
            pd.DataFrame(
                np.cumprod(1 + np.random.normal(0, 0.009, 167905)),
                index=pd.bdate_range(
                    start="2020-01-01 00:00:00+00:00",
                    end="2024-10-15 18:30:00+00:00",
                    freq=pd.offsets.Minute(15),
                ),
                columns=["ABCD"],
            ),
            3,
        ),
    ],
)
def test_intraday_momentum(
    ask_close,
    bid_close,
    close_offset_periods,
    tradingo,
):

    result = signals.intraday_momentum(
        ask_close=ask_close,
        bid_close=bid_close,
        universe=UNIVERSE,
        provider=PROVIDER,
        dry_run=True,
        close_offset_periods=close_offset_periods,
        calendar="NYSE",
        arctic=tradingo,
    )

    cal = pmc.get_calendar("NYSE")
    schedule = cal.schedule(start_date="2024-10-14", end_date="2024-10-14")
    trading_index = pmc.date_range(schedule, frequency="15min")

    subset = result.loc[trading_index]

    assert subset[-(close_offset_periods + 1) :]["signals/intraday_momentum"][
        "ABCD"
    ].to_list() == [0 for _ in range(0, 1 + close_offset_periods)]



import pandas as pd
from tradingo.cli import build_graph


PROVIDER = "test-provider"
UNIVERSE = "test-universe"
PORTFOLIO = "test-portfolio"
NAME = "test-name"


def test_build_graph():

    config = {
        "name": NAME,
        "universe": {
            UNIVERSE: {
                "provider": PROVIDER,
                "index_col": None,
                "volatility": {
                    "speeds": [32, 64],
                },
            }
        },
        "volatility": {"speeds": []},
        "signal_configs": {
            "signal1": {"function": "module.function", "args": [], "kwargs": {}},
            "signal1.capped": {
                "depends_on": ["signal1"],
                "function": "module.function",
                "args": [],
                "kwargs": {},
            },
        },
        "portfolio": {
            PORTFOLIO: {
                "function": "module.portfolio_function",
                "args": [],
                "kwargs": {
                    "signal_weights": {"signal1.capped": 1},
                },
                "universe": UNIVERSE,
                "provider": PROVIDER,
            }
        },
    }

    tasks = build_graph(config, pd.Timestamp("2018-01-01"), pd.Timestamp("2024-09-18"))

    assert tasks[PORTFOLIO].dependencies == [
        tasks[f"{UNIVERSE}.sample"],
        tasks[f"{UNIVERSE}.vol"],
        tasks[f"{UNIVERSE}.signal1.capped"],
    ]

    assert tasks[f"{PORTFOLIO}.backtest"].dependencies == [tasks[PORTFOLIO]]

    assert tasks[f"{UNIVERSE}.sample"].dependencies == [
        tasks[f"{UNIVERSE}.instruments"]
    ]
    assert tasks[f"{UNIVERSE}.signal1"].dependencies == [
        tasks[f"{UNIVERSE}.sample"],
        tasks[f"{UNIVERSE}.vol"],
    ]



import numpy as np
import pandas as pd
from pandas.core.window.rolling import window_aggregations


def omega_ratio(returns: pd.Series, required_return=0.0):
    """
    Calculate the Omega ratio of a strategy.

    :param returns: pd.Series or np.ndarray
        Daily returns of the strategy, noncumulative.

    :param required_return: float, optional
        Minimum acceptance return of the investor. Threshold over which to
        consider positive vs negative returns. It will be converted to a
        value appropriate for the period of the returns. E.g. An annual minimum
        acceptable return of 100 will translate to a minimum acceptable
        return of 0.018.

    :returns omega_ratio : float

    """
    return_threshold = (1 + required_return) ** (1 / 252) - 1
    returns_less_thresh = returns - return_threshold
    numer = sum(returns_less_thresh[returns_less_thresh > 0.0])
    denom = -1.0 * sum(returns_less_thresh[returns_less_thresh < 0.0])
    if denom > 0.0:
        return numer / denom
    else:
        return np.nan


def sharpe_ratio(returns: pd.Series, required_return=0.0):
    """
    Calculate the Sharpe ratio of a strategy.
    """
    return_threshold = (1 + required_return) ** (1 / 252) - 1
    return (returns.mean() * 252 - return_threshold) / (returns.std() * np.sqrt(252))
